#include <unistd.h>
#include <errno.h>

#define IOT_BUFSIZE 256

static int mbtk_sdl2_iot_stdio(void *tctx)
{
	sdl2_display_t *sd = tctx;
	char *bufs = malloc(IOT_BUFSIZE*2); /* can't be static because there may be multiple stdio threads for multiple displays */
	int bi;

	sd->iot_stdio.ev.type = mbtk_sdl2_ev_rstdio;
	sd->iot_stdio.ev.user.data1 = sd;

	for(bi = 0;; bi = !bi) {
		int len;
		char *buf = bi ? &bufs[IOT_BUFSIZE] : &bufs[0]; /* need to double buffer because we read in the background while the previous is processed by the user */

		len = read(0, buf, IOT_BUFSIZE-1);
		if (len > 0)
			buf[len] = '\0';
		else
			buf[0] = '\0';

		sd->iot_stdio.data = buf;
		sd->iot_stdio.len = len;

		/* wait while we are in disable */
		SDL_SemWait(sd->iot_stdio.enable);

		SDL_PushEvent(&sd->iot_stdio.ev);

		/* main thread will increase handled to 1 to indicate the event has been processed */
		SDL_SemWait(sd->iot_stdio.handled);
	}

	free(bufs);
	return 0;
}

static void mbtk_sdl2_watch_stdin_read(mbtk_display_t *disp, int enable)
{
	sdl2_display_t *sd = (sdl2_display_t *)disp;

	if (enable) {
		if (sd->iot_stdio.tid == NULL) {
			sd->iot_stdio.handled = SDL_CreateSemaphore(0);
			sd->iot_stdio.enable = SDL_CreateSemaphore(0);
			sd->iot_stdio.tid = SDL_CreateThread(mbtk_sdl2_iot_stdio, "read_stdio", sd);
		}

		/* enable reporting */
		SDL_SemTryWait(sd->iot_stdio.enable); /* decrease enable to 0 */
		SDL_SemPost(sd->iot_stdio.enable); /* increase to 1 to enable */
		sd->iot_stdio.disabled = 0;
	}
	else {
		SDL_SemTryWait(sd->iot_stdio.enable); /* decrease enable to 0 so the read thread is blocked */
		sd->iot_stdio.disabled = 1;
	}
}


/*** fdwatch ***/

typedef struct {
	int fd;
	unsigned for_read:1;
	unsigned for_write:1;
} mbtk_sdl2_fdwacth_req_t;

static int mbtk_sdl2_iot_fdwatch(void *tctx)
{
	sdl2_display_t *sd = tctx;
	mbtk_watch_fd_t wctx = {0};
	mbtk_event_t dst;

	mbtk_watch_fd_set(&wctx, sd->iot_fdwatch.ctrlr, 1, 0);
	sd->iot_fdwatch.ev.type = mbtk_sdl2_ev_fdwatch;
	sd->iot_fdwatch.ev.user.data1 = sd;

	for(;;) {
		long usec = 1000000-1;
		int res;

		res = mbtk_watch_fd_select(&wctx, &dst, &usec, NULL);
		if (res < 0) {
			if (errno == EINTR)
				continue;
			/* something went wrong with an fd, try to blocking-read control to remove stale fds */
			goto read_ctrl;
		}
		if (res == 0)
			continue;

		if (dst.data.fdwatch.fd == sd->iot_fdwatch.ctrlr) {

			mbtk_sdl2_fdwacth_req_t req;
			int len;

			read_ctrl:;

			len = read(sd->iot_fdwatch.ctrlr, &req, sizeof(req));
			if (len != sizeof(req))
				mbtk_error("mbtk_sdl2_iot_fdwatch(): short read or read error on pipe: %d != %d\n", len, (int)sizeof(req));
			else
				mbtk_watch_fd_set(&wctx, req.fd, req.for_read, req.for_write);

			continue;
		}

		sd->iot_fdwatch.fd = dst.data.fdwatch.fd;
		sd->iot_fdwatch.readable = dst.data.fdwatch.readable;
		sd->iot_fdwatch.writable = dst.data.fdwatch.writable;

		SDL_PushEvent(&sd->iot_fdwatch.ev);

		/* main thread will increase handled to 1 to indicate the event has been processed */
		SDL_SemWait(sd->iot_fdwatch.handled);
	}
	return -1; /* can't reach */
}

static void mbtk_sdl2_watch_fd(mbtk_display_t *disp, int fd, int for_read, int for_write)
{
	sdl2_display_t *sd = (sdl2_display_t *)disp;
	mbtk_sdl2_fdwacth_req_t req;
	int len;

	if (sd->iot_fdwatch.tid == NULL) {
		mbtk_be_pipe(&sd->iot_fdwatch.ctrlr, &sd->iot_fdwatch.ctrlw);
		sd->iot_fdwatch.handled = SDL_CreateSemaphore(0);
		sd->iot_fdwatch.tid = SDL_CreateThread(mbtk_sdl2_iot_fdwatch, "io_select", sd);
	}

	req.fd = fd;
	req.for_read = for_read;
	req.for_write = for_write;
	len = write(sd->iot_fdwatch.ctrlw, &req, sizeof(req)); /* lazy: assume atomic write */
	if (len != sizeof(req))
		mbtk_error("mbtk_sdl2_watch_fd(): short write or write error on pipe: %d != %d\n", len, (int)sizeof(req));
}
