#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include "../hvbox.h"
#include "../hvbox_debug.h"

#define MAX_LEVELS 32

#define get_word(str, ptr) \
do { \
	if (ptr != NULL) { \
		while(isspace(*ptr)) ptr++; \
		if (*ptr == '\0') { \
			str = ptr = NULL; \
			break; \
		} \
		str = ptr; \
		do { ptr++; } while(!isspace(*ptr) && (*ptr != '\0')); \
		if (ptr != '\0') {\
			*ptr = '\0'; \
			ptr++; \
		} \
		else \
			ptr = NULL; \
	} \
	else \
		str = NULL; \
} while(0)

#define get_int(i_out, ptr, def) \
do { \
	char *__tmp__, *__end__; \
	get_word(__tmp__, ptr);\
	if (__tmp__ != NULL) { \
		i_out = strtol(__tmp__, &__end__, 0); \
		if (*__end__ != '\0') \
			i_out = def; \
	} \
	else \
		i_out = def; \
} while(0)

#define error(...) \
do { \
	fprintf(stderr, "Error: "); \
	fprintf(stderr, __VA_ARGS__); \
	fprintf(stderr, "\n"); \
	exit(1); \
} while(0)

void dump_bfs(FILE *f, hvbox_t *root)
{
	hvbox_t *n;
	hvbox_iter_t it;

	for(n = hvbox_bfs_first(&it, root); n != NULL; n = hvbox_bfs_next(&it)) {
		hvbox_dump_node(f, it.level, n, 1);
	}
	hvbox_bfs_finish(&it);
}

int main(int argc, char *argv[])
{
	char line[1024];
	char *s, *cmd;
	hvbox_t *last[MAX_LEVELS];
	int level = 0;

	while(!(feof(stdin))) {
		*line = '\0';
		fgets(line, sizeof(line), stdin);
		s = line;
		while(isspace(*s)) s++;
		if (*s == '#')
			continue;
		for(;;) {
			get_word(cmd, s);
			if (cmd == NULL)
				break;
			if (strcmp(cmd, "root") == 0) {
				int w, h;
				get_int(w, s, -1);
				get_int(h, s, -1);
				if ((w == -1) || (h == -1))
					error("root needs integers w and h");
				last[0] = hvbox_new_root(w, h, NULL);
				level = 0;
			}
			else if (strcmp(cmd, "{") == 0) {
				level++;
				assert(level < MAX_LEVELS);
				last[level] = NULL;
			}
			else if (strcmp(cmd, "}") == 0) {
				assert(level > 0);
				level--;
			}
			else if ((strcmp(cmd, "hbox") == 0) || (strcmp(cmd, "vbox") == 0)) {
				hvbox_type_t type = (*cmd == 'h') ? HVBOX_HBOX : HVBOX_VBOX;
				assert(level > 0);
				last[level] = hvbox_append_box(last[level-1], type);
			}
			else if (strcmp(cmd, "spacer") == 0) {
				int spc;
				hvbox_type_t type;
				get_int(spc, s, -1);
				assert(level > 0);
				type = (last[level-1]->type == HVBOX_VBOX) ? HVBOX_HBOX : HVBOX_VBOX;
				last[level] = hvbox_append_box(last[level-1], type);
				if (type == HVBOX_VBOX)
					last[level]->w_min = spc;
				else
					last[level]->h_min = spc;
			}
			else if (strcmp(cmd, "widget") == 0) {
				int w, h;
				char *data;
				get_int(w, s, -1);
				get_int(h, s, -1);
				get_word(data, s);
				if ((w == -1) || (h == -1) || (data == NULL))
					error("root needs integers w, h and name");
				last[level] = hvbox_append_widget(last[level-1], w, h, strdup(data));
			}
			else if (strcmp(cmd, "align") == 0) {
				char *align;
				get_word(align, s);
				if (strcmp(align, "left") == 0) last[level-1]->align = HVBOX_LEFT_TOP;
				else if (strcmp(align, "top") == 0) last[level-1]->align = HVBOX_LEFT_TOP;
				else if (strcmp(align, "right") == 0) last[level-1]->align = HVBOX_RIGHT_BOTTOM;
				else if (strcmp(align, "bottom") == 0) last[level-1]->align = HVBOX_RIGHT_BOTTOM;
				else if (strcmp(align, "justify") == 0) last[level-1]->align = HVBOX_JUSTIFY;
				else if (strcmp(align, "center") == 0) last[level-1]->align = HVBOX_CENTER;
				else error("Unknown alignment: '%s'", align);
			}
			else if (strcmp(cmd, "span") == 0) {
				char *spans;
				hvbox_span_t span;
				get_word(spans, s);
				if (strcmp(spans, "minimal") == 0) span = HVBOX_MINIMAL;
				else if (strcmp(spans, "fill") == 0) span = HVBOX_FILL;
				else {
					span = atoi(spans);
					if ((span < 0) || (span > 100))
						error("Invalid integer span: '%s'", spans);
				}
				last[level-1]->span = span;
			}
			else if (strcmp(cmd, "padding") == 0) {
				int pad;
				get_int(pad, s, -1);
				assert(level > 0);
				if (pad < 0)
					error("invalid padding value");
				last[level-1]->padding = pad;
			}
			else if (strcmp(cmd, "spacing") == 0) {
				int spc;
				get_int(spc, s, -1);
				assert(level > 0);
				if (spc < 0)
					error("invalid padding value");
				last[level-1]->spacing = spc;
			}
			else if (strncmp(cmd, "fr_", 3) == 0) {
				int fr;
				get_int(fr, s, -1);
				assert(level > 0);
				if ((fr < 0) || (fr > 255))
					error("invalid fr_ value");
				switch(cmd[3]) {
					case 'l': last[level-1]->fr_l = fr; break;
					case 'r': last[level-1]->fr_r = fr; break;
					case 't': last[level-1]->fr_t = fr; break;
					case 'b': last[level-1]->fr_b = fr; break;
					default: error("invalid fr_ type: must be fr_l, fr_r, fr_t or fr_b");
				}
			}
		}
	}

	if (level != 0)
		error("Not enough closing brakcets\n");

	hvbox_update(last[0]);

	hvbox_dump_nodes(stdout, 0, last[0], 1);

	if (argc > 1) {
		printf("\nBFS:\n");
		dump_bfs(stdout, last[0]);
	}

	return 0;
}
